be much more careful about sending values backwards...
[model-checker.git] / model.h
diff --git a/model.h b/model.h
index 4b9e3ff53f0082b30efd095b5618186480882d5e..c99e0f8b424ddbe8079593713e498e53fe6fb80e 100644 (file)
--- a/model.h
+++ b/model.h
 #include <cstddef>
 #include <ucontext.h>
 
-#include "schedule.h"
 #include "mymemory.h"
-#include "libthreads.h"
-#include "threads.h"
 #include "action.h"
-#include "clockvector.h"
 #include "hashtable.h"
 #include "workqueue.h"
+#include "config.h"
+#include "modeltypes.h"
 
 /* Forward declaration */
 class NodeStack;
 class CycleGraph;
 class Promise;
+class Scheduler;
+class Thread;
 
 /** @brief Shorthand for a list of release sequence heads */
-typedef std::vector< const ModelAction *, MyAlloc<const ModelAction *> > rel_heads_list_t;
+typedef std::vector< const ModelAction *, ModelAlloc<const ModelAction *> > rel_heads_list_t;
 
 /**
  * Model checker parameter structure. Holds run-time configuration options for
@@ -39,8 +39,7 @@ struct model_params {
 };
 
 struct PendingFutureValue {
-       uint64_t value;
-       modelclock_t expiration;
+       ModelAction *writer;
        ModelAction * act;
 };
 
@@ -49,12 +48,25 @@ struct PendingFutureValue {
  */
 struct model_snapshot_members {
        ModelAction *current_action;
-       int next_thread_id;
+       unsigned int next_thread_id;
        modelclock_t used_sequence_numbers;
        Thread *nextThread;
        ModelAction *next_backtrack;
 };
 
+/** @brief Records information regarding a single pending release sequence */
+struct release_seq {
+       /** @brief The acquire operation */
+       ModelAction *acquire;
+       /** @brief The head of the RMW chain from which 'acquire' reads; may be
+        *  equal to 'release' */
+       const ModelAction *rf;
+       /** @brief The head of the potential longest release sequence chain */
+       const ModelAction *release;
+       /** @brief The write(s) that may break the release sequence */
+       std::vector<const ModelAction *> writes;
+};
+
 /** @brief The central structure for model-checking */
 class ModelChecker {
 public:
@@ -66,18 +78,18 @@ public:
 
        /** Prints an execution summary with trace information. */
        void print_summary();
+#if SUPPORT_MOD_ORDER_DUMP
+       void dumpGraph(char *filename);
+#endif
 
        void add_thread(Thread *t);
        void remove_thread(Thread *t);
-       Thread * get_thread(thread_id_t tid) { return thread_map->get(id_to_int(tid)); }
-       Thread * get_thread(ModelAction *act) { return get_thread(act->get_tid()); }
+       Thread * get_thread(thread_id_t tid) const;
+       Thread * get_thread(ModelAction *act) const;
 
        thread_id_t get_next_id();
-       int get_num_threads();
-       modelclock_t get_next_seq_num();
-
-       /** @return The currently executing Thread. */
-       Thread * get_current_thread() { return scheduler->get_current_thread(); }
+       unsigned int get_num_threads();
+       Thread * get_current_thread();
 
        int switch_to_master(ModelAction *act);
        ClockVector * get_cv(thread_id_t tid);
@@ -86,7 +98,8 @@ public:
        bool isfeasible();
        bool isfeasibleotherthanRMW();
        bool isfinalfeasible();
-       void check_promises(ClockVector *old_cv, ClockVector * merge_cv);
+       void mo_check_promises(thread_id_t tid, const ModelAction *write);
+       void check_promises(thread_id_t tid, ClockVector *old_cv, ClockVector * merge_cv);
        void get_release_seq_heads(ModelAction *act, rel_heads_list_t *release_heads);
        void finish_execution();
        bool isfeasibleprefix();
@@ -97,18 +110,24 @@ public:
        void set_bad_synchronization() { bad_synchronization = true; }
 
        const model_params params;
+       Scheduler * get_scheduler() { return scheduler;}
 
        MEMALLOC
 private:
        /** The scheduler to use: tracks the running/ready Threads */
        Scheduler *scheduler;
 
+       bool sleep_can_read_from(ModelAction * curr, const ModelAction *write);
        bool thin_air_constraint_may_allow(const ModelAction * writer, const ModelAction *reader);
+       bool mo_may_allow(const ModelAction * writer, const ModelAction *reader);
        bool has_asserted() {return asserted;}
        void reset_asserted() {asserted=false;}
        int num_executions;
        int num_feasible_executions;
        bool promises_expired();
+       void execute_sleep_set();
+       void wake_up_sleeping_actions(ModelAction * curr);
+       modelclock_t get_next_seq_num();
 
        /**
         * Stores the ModelAction for the current thread action.  Call this
@@ -123,6 +142,7 @@ private:
        bool process_write(ModelAction *curr);
        bool process_mutex(ModelAction *curr);
        bool process_thread_action(ModelAction *curr);
+       void process_relseq_fixup(ModelAction *curr, work_queue_t *work_queue);
        bool check_action_enabled(ModelAction *curr);
 
        bool take_step();
@@ -135,6 +155,7 @@ private:
        void reset_to_initial_state();
        bool resolve_promises(ModelAction *curr);
        void compute_promises(ModelAction *curr);
+       void compute_relseq_breakwrites(ModelAction *curr);
 
        void check_curr_backtracking(ModelAction * curr);
        void add_action_to_lists(ModelAction *act);
@@ -146,11 +167,12 @@ private:
        void post_r_modification_order(ModelAction *curr, const ModelAction *rf);
        bool r_modification_order(ModelAction *curr, const ModelAction *rf);
        bool w_modification_order(ModelAction *curr);
-       bool release_seq_head(const ModelAction *rf, rel_heads_list_t *release_heads) const;
+       bool release_seq_heads(const ModelAction *rf, rel_heads_list_t *release_heads, struct release_seq *pending) const;
        bool resolve_release_sequences(void *location, work_queue_t *work_queue);
        void do_complete_join(ModelAction *join);
 
        ModelAction *diverge;
+       ModelAction *earliest_diverge;
 
        ucontext_t system_context;
        action_list_t *action_trace;
@@ -165,24 +187,28 @@ private:
        HashTable<const void *, action_list_t, uintptr_t, 4> *lock_waiters_map;
 
        HashTable<void *, std::vector<action_list_t>, uintptr_t, 4 > *obj_thrd_map;
-       std::vector<Promise *> *promises;
-       std::vector<struct PendingFutureValue> *futurevalues;
+       std::vector< Promise *, SnapshotAlloc<Promise *> > *promises;
+       std::vector< struct PendingFutureValue, SnapshotAlloc<struct PendingFutureValue> > *futurevalues;
 
        /**
* List of acquire actions that might synchronize with one or more
* release sequence. Release sequences might be determined lazily as
* promises are fulfilled and modification orders are established. Each
* ModelAction in this list must be an acquire operation.
         */
-       std::vector<ModelAction *> *pending_acq_rel_seq;
+       std::vector< struct release_seq *, SnapshotAlloc<struct release_seq *> > *pending_rel_seqs;
 
-       std::vector<ModelAction *> *thrd_last_action;
+       std::vector< ModelAction *, SnapshotAlloc<ModelAction *> > *thrd_last_action;
        NodeStack *node_stack;
 
        /** Private data members that should be snapshotted. They are grouped
         * together for efficiency and maintainability. */
        struct model_snapshot_members *priv;
 
+       /** A special model-checker Thread; used for associating with
+        *  model-checker-related ModelAcitons */
+       Thread *model_thread;
+
        /**
         * @brief The modification order graph
         *
@@ -165,24 +187,28 @@ private:
 
        /**
-        * List of acquire actions that might synchronize with one or more
-        * release sequence. Release sequences might be determined lazily as
-        * promises are fulfilled and modification orders are established. Each
-        * ModelAction in this list must be an acquire operation.
+        * List of pending release sequences. Release sequences might be
+        * determined lazily as promises are fulfilled and modification orders
+        * are established. Each entry in the list may only be partially
+        * filled, depending on its pending status.
         */